# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: transact.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from . import types_pb2 as types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0etransact.proto\x12\x10\x61\x65rospike.vector\x1a\x1bgoogle/protobuf/empty.proto\x1a\x0btypes.proto\"U\n\nPutRequest\x12\"\n\x03key\x18\x01 \x01(\x0b\x32\x15.aerospike.vector.Key\x12#\n\x04\x62ins\x18\x03 \x03(\x0b\x32\x15.aerospike.vector.Bin\"d\n\nGetRequest\x12\"\n\x03key\x18\x01 \x01(\x0b\x32\x15.aerospike.vector.Key\x12\x32\n\x0b\x62inSelector\x18\x02 \x01(\x0b\x32\x1d.aerospike.vector.BinSelector\"b\n\x10IsIndexedRequest\x12\"\n\x03key\x18\x01 \x01(\x0b\x32\x15.aerospike.vector.Key\x12*\n\x07indexId\x18\x02 \x01(\x0b\x32\x19.aerospike.vector.IndexId\"P\n\x0b\x42inSelector\x12/\n\x04type\x18\x01 \x01(\x0e\x32!.aerospike.vector.BinSelectorType\x12\x10\n\x08\x62inNames\x18\x02 \x03(\t\"\x81\x02\n\x13VectorSearchRequest\x12(\n\x05index\x18\x01 \x01(\x0b\x32\x19.aerospike.vector.IndexId\x12-\n\x0bqueryVector\x18\x02 \x01(\x0b\x32\x18.aerospike.vector.Vector\x12\r\n\x05limit\x18\x03 \x01(\r\x12\x32\n\x0b\x62inSelector\x18\x04 \x01(\x0b\x32\x1d.aerospike.vector.BinSelector\x12>\n\x10hnswSearchParams\x18\x05 \x01(\x0b\x32\".aerospike.vector.HnswSearchParamsH\x00\x42\x0e\n\x0csearchParams*3\n\x0f\x42inSelectorType\x12\x07\n\x03\x41LL\x10\x00\x12\x08\n\x04NONE\x10\x01\x12\r\n\tSPECIFIED\x10\x02\x32\xed\x02\n\x08Transact\x12=\n\x03Put\x12\x1c.aerospike.vector.PutRequest\x1a\x16.google.protobuf.Empty\"\x00\x12?\n\x03Get\x12\x1c.aerospike.vector.GetRequest\x1a\x18.aerospike.vector.Record\"\x00\x12<\n\x06\x45xists\x12\x15.aerospike.vector.Key\x1a\x19.aerospike.vector.Boolean\"\x00\x12L\n\tIsIndexed\x12\".aerospike.vector.IsIndexedRequest\x1a\x19.aerospike.vector.Boolean\"\x00\x12U\n\x0cVectorSearch\x12%.aerospike.vector.VectorSearchRequest\x1a\x1a.aerospike.vector.Neighbor\"\x00\x30\x01\x42=\n\x1b\x63om.aerospike.vector.clientP\x01Z\x1c\x61\x65rospike.com/vector/protos/b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'transact_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\033com.aerospike.vector.clientP\001Z\034aerospike.com/vector/protos/'
  _globals['_BINSELECTORTYPE']._serialized_start=709
  _globals['_BINSELECTORTYPE']._serialized_end=760
  _globals['_PUTREQUEST']._serialized_start=78
  _globals['_PUTREQUEST']._serialized_end=163
  _globals['_GETREQUEST']._serialized_start=165
  _globals['_GETREQUEST']._serialized_end=265
  _globals['_ISINDEXEDREQUEST']._serialized_start=267
  _globals['_ISINDEXEDREQUEST']._serialized_end=365
  _globals['_BINSELECTOR']._serialized_start=367
  _globals['_BINSELECTOR']._serialized_end=447
  _globals['_VECTORSEARCHREQUEST']._serialized_start=450
  _globals['_VECTORSEARCHREQUEST']._serialized_end=707
  _globals['_TRANSACT']._serialized_start=763
  _globals['_TRANSACT']._serialized_end=1128
# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from . import vector_db_pb2 as vector__db__pb2


class AboutStub(object):
    """Information about the service.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Get = channel.unary_unary(
                '/aerospike.vector.About/Get',
                request_serializer=vector__db__pb2.AboutRequest.SerializeToString,
                response_deserializer=vector__db__pb2.AboutResponse.FromString,
                )


class AboutServicer(object):
    """Information about the service.
    """

    def Get(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_AboutServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Get': grpc.unary_unary_rpc_method_handler(
                    servicer.Get,
                    request_deserializer=vector__db__pb2.AboutRequest.FromString,
                    response_serializer=vector__db__pb2.AboutResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'aerospike.vector.About', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class About(object):
    """Information about the service.
    """

    @staticmethod
    def Get(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/aerospike.vector.About/Get',
            vector__db__pb2.AboutRequest.SerializeToString,
            vector__db__pb2.AboutResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)


class ClusterInfoStub(object):
    """Vector DB cluster service.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetNodeId = channel.unary_unary(
                '/aerospike.vector.ClusterInfo/GetNodeId',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=vector__db__pb2.NodeId.FromString,
                )
        self.GetClusterId = channel.unary_unary(
                '/aerospike.vector.ClusterInfo/GetClusterId',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=vector__db__pb2.ClusterId.FromString,
                )
        self.GetClusterEndpoints = channel.unary_unary(
                '/aerospike.vector.ClusterInfo/GetClusterEndpoints',
                request_serializer=vector__db__pb2.ClusterNodeEndpointsRequest.SerializeToString,
                response_deserializer=vector__db__pb2.ClusterNodeEndpoints.FromString,
                )
        self.GetOwnedPartitions = channel.unary_unary(
                '/aerospike.vector.ClusterInfo/GetOwnedPartitions',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=vector__db__pb2.ClusterPartitions.FromString,
                )


class ClusterInfoServicer(object):
    """Vector DB cluster service.
    """

    def GetNodeId(self, request, context):
        """Get the internal cluster node-Id for this server.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetClusterId(self, request, context):
        """Get current cluster-Id for the current cluster.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetClusterEndpoints(self, request, context):
        """Get the advertised/listening endpoints for all nodes in the cluster, given a listener name.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetOwnedPartitions(self, request, context):
        """Get per-node owned partition list for all nodes in the cluster.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ClusterInfoServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetNodeId': grpc.unary_unary_rpc_method_handler(
                    servicer.GetNodeId,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=vector__db__pb2.NodeId.SerializeToString,
            ),
            'GetClusterId': grpc.unary_unary_rpc_method_handler(
                    servicer.GetClusterId,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=vector__db__pb2.ClusterId.SerializeToString,
            ),
            'GetClusterEndpoints': grpc.unary_unary_rpc_method_handler(
                    servicer.GetClusterEndpoints,
                    request_deserializer=vector__db__pb2.ClusterNodeEndpointsRequest.FromString,
                    response_serializer=vector__db__pb2.ClusterNodeEndpoints.SerializeToString,
            ),
            'GetOwnedPartitions': grpc.unary_unary_rpc_method_handler(
                    servicer.GetOwnedPartitions,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=vector__db__pb2.ClusterPartitions.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'aerospike.vector.ClusterInfo', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ClusterInfo(object):
    """Vector DB cluster service.
    """

    @staticmethod
    def GetNodeId(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/aerospike.vector.ClusterInfo/GetNodeId',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            vector__db__pb2.NodeId.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetClusterId(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/aerospike.vector.ClusterInfo/GetClusterId',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            vector__db__pb2.ClusterId.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetClusterEndpoints(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/aerospike.vector.ClusterInfo/GetClusterEndpoints',
            vector__db__pb2.ClusterNodeEndpointsRequest.SerializeToString,
            vector__db__pb2.ClusterNodeEndpoints.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetOwnedPartitions(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/aerospike.vector.ClusterInfo/GetOwnedPartitions',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            vector__db__pb2.ClusterPartitions.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import asyncio
import logging
import sys
import time
from typing import Any, Optional, Union

import google.protobuf.empty_pb2
import grpc

from . import conversions
from . import index_pb2
from . import index_pb2_grpc
from . import transact_pb2
from . import transact_pb2_grpc
from . import types
from . import types_pb2
from . import vectordb_channel_provider

empty = google.protobuf.empty_pb2.Empty()
logger = logging.getLogger(__name__)


class VectorDbClient(object):
    """
    Aerospike Vector Vector Client

    This client specializes in performing database operations with vector data.
    Moreover, the client supports Hierarchical Navigable Small World (HNSW) vector searches,
    allowing users to find vectors similar to a given query vector within an index.
    """

    def __init__(
        self,
        *,
        seeds: Union[types.HostPort, tuple[types.HostPort, ...]],
        listener_name: str = None,
        is_loadbalancer: Optional[bool] = False,
    ) -> None:
        """
        Initialize the Aerospike Vector Vector Client.

        Args:
            seeds (Union[types.HostPort, tuple[types.HostPort, ...]]):
                Used to create appropriate gRPC channels for interacting with Aerospike Vector.
            listener_name (Optional[str], optional):
                Advertised listener for the client. Defaults to None.
            is_loadbalancer (bool, optional):
                If true, the first seed address will be treated as a load balancer node.

        Raises:
            Exception: Raised when no seed host is provided.
        """
        if not seeds:
            raise Exception("at least one seed host needed")

        if isinstance(seeds, types.HostPort):
            seeds = (seeds,)

        self._channelProvider = vectordb_channel_provider.VectorDbChannelProvider(
            seeds, listener_name, is_loadbalancer
        )

    async def put(
        self,
        *,
        namespace: str,
        key: Union[int, str, bytes, bytearray],
        record_data: dict[str, Any],
        set_name: Optional[str] = None,
    ) -> None:
        """
        Write a record to Aerospike Vector.

        Args:
            namespace (str): The namespace for the record.
            key (Union[int, str, bytes, bytearray]): The key for the record.
            record_data (dict[str, Any]): The data to be stored in the record.
            set_name (Optional[str], optional): The name of the set to which the record belongs. Defaults to None.

        Raises:
            grpc.RpcError: Raised if an error occurs during the RPC communication with the server while attempting to create the index.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.

        """
        transact_stub = transact_pb2_grpc.TransactStub(
            self._channelProvider.get_channel()
        )
        key = _get_key(namespace, set_name, key)
        bin_list = [
            types_pb2.Bin(name=k, value=conversions.toVectorDbValue(v))
            for (k, v) in record_data.items()
        ]
        logger.debug(
            "Putting record: namespace=%s, key=%s, record_data:%s, set_name:%s",
            namespace,
            key,
            record_data,
            set_name,
        )
        try:
            await transact_stub.Put(transact_pb2.PutRequest(key=key, bins=bin_list))
        except grpc.RpcError as e:
            logger.error("Failed with error: %s", e)
            raise e

    async def get(
        self,
        *,
        namespace: str,
        key: Union[int, str, bytes, bytearray],
        bin_names: Optional[list[str]] = None,
        set_name: Optional[str] = None,
    ) -> types.RecordWithKey:
        """
        Read a record from Aerospike Vector.

        Args:
            namespace (str): The namespace for the record.
            key (Union[int, str, bytes, bytearray]): The key for the record.
            bin_names (Optional[list[str]], optional): A list of bin names to retrieve from the record.
            If None, all bins are retrieved. Defaults to None.
            set_name (Optional[str], optional): The name of the set from which to read the record. Defaults to None.

        Returns:
            types.RecordWithKey: A record with its associated key.

        Raises:
            grpc.RpcError: Raised if an error occurs during the RPC communication with the server while attempting to create the index.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.
        """
        transact_stub = transact_pb2_grpc.TransactStub(
            self._channelProvider.get_channel()
        )
        key = _get_key(namespace, set_name, key)
        bin_selector = _get_bin_selector(bin_names=bin_names)
        logger.debug(
            "Getting record: namespace=%s, key=%s, bin_names:%s, set_name:%s",
            namespace,
            key,
            bin_names,
            set_name,
        )
        try:
            result = await transact_stub.Get(
                transact_pb2.GetRequest(key=key, binSelector=bin_selector)
            )
        except grpc.RpcError as e:
            logger.error("Failed with error: %s", e)
            raise e

        return types.RecordWithKey(
            key=conversions.fromVectorDbKey(key),
            bins=conversions.fromVectorDbRecord(result),
        )

    async def exists(
        self, *, namespace: str, key: Any, set_name: Optional[str] = None
    ) -> bool:
        """
        Check if a record exists in Aerospike Vector.

        Args:
            namespace (str): The namespace for the record.
            key (Any): The key for the record.
            set_name (Optional[str], optional): The name of the set to which the record belongs. Defaults to None.

        Returns:
            bool: True if the record exists, False otherwise.

        Raises:
            grpc.RpcError: Raised if an error occurs during the RPC communication with the server while attempting to create the index.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.
        """
        transact_stub = transact_pb2_grpc.TransactStub(
            self._channelProvider.get_channel()
        )
        key = _get_key(namespace, set_name, key)
        logger.debug(
            "Getting record existence: namespace=%s, key=%s, set_name:%s",
            namespace,
            key,
            set_name,
        )
        try:
            result = await transact_stub.Exists(key)
        except grpc.RpcError as e:
            logger.error("Failed with error: %s", e)
            raise e

        return result.value

    async def is_indexed(
        self,
        *,
        namespace: str,
        key: Union[int, str, bytes, bytearray],
        index_name: str,
        index_namespace: Optional[str] = None,
        set_name: Optional[str] = None,
    ) -> bool:
        """
        Check if a record is indexed in the Vector DB.

        Args:
            namespace (str): The namespace for the record.
            key (Union[int, str, bytes, bytearray]): The key for the record.
            index_name (str): The name of the index.
            index_namespace (Optional[str], optional): The namespace of the index.
            If None, defaults to the namespace of the record. Defaults to None.
            set_name (Optional[str], optional): The name of the set to which the record belongs. Defaults to None.

        Returns:
            bool: True if the record is indexed, False otherwise.

        Raises:
            grpc.RpcError: Raised if an error occurs during the RPC communication with the server while attempting to create the index.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.
        """
        if not index_namespace:
            index_namespace = namespace

        index_id = types_pb2.IndexId(namespace=index_namespace, name=index_name)
        key = _get_key(namespace, set_name, key)
        request = transact_pb2.IsIndexedRequest(key=key, indexId=index_id)
        transact_stub = transact_pb2_grpc.TransactStub(
            self._channelProvider.get_channel()
        )
        logger.debug(
            "Checking if index exists: namespace=%s, key=%s, index_name=%s, index_namespace=%s, set_name=%s",
            namespace,
            key,
            index_name,
            index_namespace,
            set_name,
        )
        try:
            result = await transact_stub.IsIndexed(request)
        except grpc.RpcError as e:
            logger.error("Failed with error: %s", e)
            raise e
        return result.value

    async def vector_search(
        self,
        *,
        namespace: str,
        index_name: str,
        query: list[Union[bool, float]],
        limit: int,
        search_params: Optional[types.HnswSearchParams] = None,
        bin_names: Optional[list[str]] = None,
    ) -> list[types.Neighbor]:
        """
        Perform a Hierarchical Navigable Small World (HNSW) vector search in Aerospike Vector.

        Args:
            namespace (str): The namespace for the records.
            index_name (str): The name of the index.
            query (list[Union[bool, float]]): The query vector for the search.
            limit (int): The maximum number of neighbors to return. K value.
            search_params (Optional[types_pb2.HnswSearchParams], optional): Parameters for the HNSW algorithm.
            If None, the default parameters for the index are used. Defaults to None.
            bin_names (Optional[list[str]], optional): A list of bin names to retrieve from the results.
            If None, all bins are retrieved. Defaults to None.

        Returns:
            list[types.Neighbor]: A list of neighbors records found by the search.

        Raises:
            grpc.RpcError: Raised if an error occurs during the RPC communication with the server while attempting to create the index.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.
        """
        transact_stub = transact_pb2_grpc.TransactStub(
            self._channelProvider.get_channel()
        )
        logger.debug(
            "Performing vector search: namespace=%s, index_name=%s, query=%s, limit=%s, search_params=%s, bin_names=%s",
            namespace,
            index_name,
            query,
            limit,
            search_params,
            bin_names,
        )
        if search_params != None:
            search_params = search_params._to_pb2()
        try:
            results_stream = transact_stub.VectorSearch(
                transact_pb2.VectorSearchRequest(
                    index=types_pb2.IndexId(namespace=namespace, name=index_name),
                    queryVector=(conversions.toVectorDbValue(query).vectorValue),
                    limit=limit,
                    hnswSearchParams=search_params,
                    binSelector=_get_bin_selector(bin_names=bin_names),
                )
            )
        except grpc.RpcError as e:
            logger.error("Failed with error: %s", e)
            raise e
        async_results = []
        async for result in results_stream:
            async_results.append(conversions.fromVectorDbNeighbor(result))

        return async_results

    async def wait_for_index_completion(
        self, *, namespace: str, name: str, timeout: Optional[int] = sys.maxsize
    ) -> None:
        """
        Wait for the index to have no pending index update operations.

        Args:
            namespace (str): The namespace of the index.
            name (str): The name of the index.
            timeout (int, optional): The maximum time (in seconds) to wait for the index to complete.
            Defaults to sys.maxsize.

        Raises:
            Exception: Raised when the timeout occurs while waiting for index completion.
            grpc.RpcError: Raised if an error occurs during the RPC communication with the server while attempting to create the index.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.

        Note:
            The function polls the index status with a wait interval of 10 seconds until either
            the timeout is reached or the index has no pending index update operations.
        """
        # Wait interval between polling
        index_stub = index_pb2_grpc.IndexServiceStub(
            self._channelProvider.get_channel()
        )

        wait_interval = 1

        unmerged_record_count = sys.maxsize
        start_time = time.monotonic()
        unmerged_record_initialized = False
        while True:
            if start_time + timeout < time.monotonic():
                raise "timed-out waiting for index completion"
            # Wait for in-memory batches to be flushed to storage.
            if start_time + 10 < time.monotonic():
                unmerged_record_initialized = True
            try:
                index_status = await index_stub.GetStatus(
                    types_pb2.IndexId(namespace=namespace, name=name)
                )

            except grpc.RpcError as e:
                if e.code() == grpc.StatusCode.UNAVAILABLE:
                    continue
                else:
                    logger.error("Failed with error: %s", e)
                    raise e

            if index_status.unmergedRecordCount > 0:
                unmerged_record_initialized = True

            if (
                unmerged_record_count == 0
                and index_status.unmergedRecordCount == 0
                and unmerged_record_initialized == True
            ):
                return
            # Update.
            unmerged_record_count = index_status.unmergedRecordCount
            await asyncio.sleep(wait_interval)

    async def close(self):
        """
        Close the Aerospike Vector Vector Client.

        This method closes gRPC channels connected to Aerospike Vector.

        Note:
            This method should be called when the VectorDbAdminClient is no longer needed to release resources.
        """
        await self._channelProvider.close()

    async def __aenter__(self):
        """
        Enter an asynchronous context manager for the vector client.

        Returns:
            VectorDbClient: Aerospike Vector Vector Client instance.
        """
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Exit an asynchronous context manager for the vector client.
        """
        await self.close()


def _get_bin_selector(*, bin_names: Optional[list] = None):

    if not bin_names:
        bin_selector = transact_pb2.BinSelector(
            type=transact_pb2.BinSelectorType.ALL, binNames=bin_names
        )
    else:
        bin_selector = transact_pb2.BinSelector(
            type=transact_pb2.BinSelectorType.SPECIFIED, binNames=bin_names
        )
    return bin_selector


def _get_key(namespace: str, set: str, key: Union[int, str, bytes, bytearray]):
    if isinstance(key, str):
        key = types_pb2.Key(namespace=namespace, set=set, stringValue=key)
    elif isinstance(key, int):
        key = types_pb2.Key(namespace=namespace, set=set, longValue=key)
    elif isinstance(key, (bytes, bytearray)):
        key = types_pb2.Key(namespace=namespace, set=set, bytesValue=key)
    else:
        raise Exception("Invalid key type" + type(key))
    return key



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: index.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from . import types_pb2 as types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0bindex.proto\x12\x10\x61\x65rospike.vector\x1a\x1bgoogle/protobuf/empty.proto\x1a\x0btypes.proto\"2\n\x13IndexStatusResponse\x12\x1b\n\x13unmergedRecordCount\x18\x02 \x01(\x03\x32\xf3\x02\n\x0cIndexService\x12\x45\n\x06\x43reate\x12!.aerospike.vector.IndexDefinition\x1a\x16.google.protobuf.Empty\"\x00\x12;\n\x04\x44rop\x12\x19.aerospike.vector.IndexId\x1a\x16.google.protobuf.Empty\"\x00\x12G\n\x04List\x12\x16.google.protobuf.Empty\x1a%.aerospike.vector.IndexDefinitionList\"\x00\x12\x45\n\x03Get\x12\x19.aerospike.vector.IndexId\x1a!.aerospike.vector.IndexDefinition\"\x00\x12O\n\tGetStatus\x12\x19.aerospike.vector.IndexId\x1a%.aerospike.vector.IndexStatusResponse\"\x00\x42=\n\x1b\x63om.aerospike.vector.clientP\x01Z\x1c\x61\x65rospike.com/vector/protos/b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'index_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\033com.aerospike.vector.clientP\001Z\034aerospike.com/vector/protos/'
  _globals['_INDEXSTATUSRESPONSE']._serialized_start=75
  _globals['_INDEXSTATUSRESPONSE']._serialized_end=125
  _globals['_INDEXSERVICE']._serialized_start=128
  _globals['_INDEXSERVICE']._serialized_end=499
# @@protoc_insertion_point(module_scope)



import enum
from typing import Any, Optional

from . import types_pb2


class HostPort(object):
    """
    represents host, port and TLS usage information.
    Used primarily when intializing client.

    Args:
        host (str): The host address.
        port (int): The port number.
        isTls (Optional[bool], optional): Indicates if TLS is enabled. Defaults to False.
    """

    def __init__(self, *, host: str, port: int, isTls: Optional[bool] = False) -> None:
        self.host = host
        self.port = port
        self.isTls = isTls


class Key(object):
    """
    Represents a record key.
    Used in RecordWithKey.

    Args:
        namespace (str): The namespace for the key.
        set (str): The set for the key.
        digest (bytearray): The digest of the key.
        key (Any): The key itself.
    """

    def __init__(
        self, *, namespace: str, set: str, digest: bytearray, key: Any
    ) -> None:
        self.namespace = namespace
        self.set = set
        self.digest = digest
        self.key = key

    def __str__(self):
        """
        Returns a string representation of the key.
        """
        return f"Key: namespace='{self.namespace}', set='{self.set}', digest={self.digest}, key={self.key}"


class RecordWithKey(object):
    """
    Represents a record, including a key and bins.
    Return value for VectorDbClient.get.

    Args:
        key (Key): The key of the record.
        bins (dict[str, Any]): The bins associated with the record.
    """

    def __init__(self, *, key: Key, bins: dict[str, Any]) -> None:
        self.key = key
        self.bins = bins

    def __str__(self):
        """
        Returns a string representation of the record, including a key and bins.
        """
        bins_info = ""
        for key, value in self.bins.items():
            if isinstance(value, list):
                if len(value) > 4:
                    value_str = (
                        "[\n"
                        + ",\n".join("\t\t\t{}".format(val) for val in value[:3])
                        + ",\n\t\t\t...\n\t\t]"
                    )
                else:
                    value_str = str(value)
            else:
                value_str = str(value)
            bins_info += "\n\t\t{}: {}".format(key, value_str)
        return "{{\n\t{},\n\tbins: {{\n{}\n\t}}\n}}".format(self.key, bins_info)


class Neighbor(object):
    """
    Represents a neighboring record in the context of approximate nearest neighbor search.

    This class represents a neighboring record in relation to a query record. It includes information such as the key, bins, and distance from the query record.

    Args:
        key (Key): The Key instance identifying the neighboring record.
        bins (dict[str, Any]): A dictionary representing bins associated with the neighboring record.
        distance (float): The distance between the neighboring record and the query record, calculated based on the chosen VectorDistanceMetric.

    Notes:
        - The distance metric used to calculate the distance between records is determined by the chosen VectorDistanceMetric.
        - The neighbor's distance indicates how similar or dissimilar it is to the query record based on the chosen distance metric.
        - A smaller distance typically implies greater similarity between records.

    """

    def __init__(self, *, key: Key, bins: dict[str, Any], distance: float) -> None:
        self.key = key
        self.bins = bins
        self.distance = distance

    def __str__(self):
        """
        Returns a string representation of the neighboring record.
        """
        bins_info = ""
        for key, value in self.bins.items():
            if isinstance(value, list):
                if len(value) > 4:
                    value_str = (
                        "[\n"
                        + ",\n".join("\t\t\t{}".format(val) for val in value[:3])
                        + ",\n\t\t\t...\n\t\t]"
                    )
                else:
                    value_str = str(value)
            else:
                value_str = str(value)
            bins_info += "\n\t\t{}: {}".format(key, value_str)
        return "{{\n\t{},\n\tdistance: {},\n\tbins: {{\n{}\n\t}}\n}}".format(
            self.key, self.distance, bins_info
        )


class VectorDistanceMetric(enum.Enum):
    """
    Enumeration of vector distance metrics.
    """

    SQUARED_EUCLIDEAN: types_pb2.VectorDistanceMetric = (
        types_pb2.VectorDistanceMetric.SQUARED_EUCLIDEAN
    )
    COSINE: types_pb2.VectorDistanceMetric = types_pb2.VectorDistanceMetric.COSINE
    DOT_PRODUCT: types_pb2.VectorDistanceMetric = (
        types_pb2.VectorDistanceMetric.DOT_PRODUCT
    )
    MANHATTAN: types_pb2.VectorDistanceMetric = types_pb2.VectorDistanceMetric.MANHATTAN
    HAMMING: types_pb2.VectorDistanceMetric = types_pb2.VectorDistanceMetric.HAMMING


class HnswBatchingParams(object):
    """
    Parameters for configuring batching behaviour for batch based index update.

    Args:
        max_records (Optional[int], optional): Maximum number of records to fit in a batch. Defaults to 10000.
        interval (Optional[int], optional): The maximum amount of time in milliseconds to wait before finalizing a batch. Defaults to 10000.
        disabled (Optional[bool], optional): Disables batching for index updates. Default is False.
    """

    def __init__(
        self,
        *,
        max_records: Optional[int] = 10000,
        interval: Optional[int] = 10000,
        disabled: Optional[bool] = False,
    ) -> None:
        self.max_records = max_records
        self.interval = interval
        self.disabled = disabled

    def _to_pb2(self):
        params = types_pb2.HnswBatchingParams()
        params.maxRecords = self.max_records
        params.interval = self.interval
        params.disabled = self.disabled
        return params


class HnswParams(object):
    """
    Parameters for the Hierarchical Navigable Small World (HNSW) algorithm, used for approximate nearest neighbor search.

    Args:
        m (Optional[int], optional): The number of bi-directional links created per level during construction. Larger 'm' values lead to higher recall but slower construction. Defaults to 16.
        ef_construction (Optional[int], optional): The size of the dynamic list for the nearest neighbors (candidates) during the index construction. Larger 'ef_construction' values lead to higher recall but slower construction. Defaults to 100.
        ef (Optional[int], optional): The size of the dynamic list for the nearest neighbors (candidates) during the search phase. Larger 'ef' values lead to higher recall but slower search. Defaults to 100.
        batching_params (Optional[HnswBatchingParams], optional): Parameters related to configuring batch processing, such as the maximum number of records per batch and batching interval. Defaults to HnswBatchingParams().
    """

    def __init__(
        self,
        *,
        m: Optional[int] = 16,
        ef_construction: Optional[int] = 100,
        ef: Optional[int] = 100,
        batching_params: Optional[HnswBatchingParams] = HnswBatchingParams(),
    ) -> None:
        self.m = m
        self.ef_construction = ef_construction
        self.ef = ef
        self.batching_params = batching_params

    def _to_pb2(self):
        params = types_pb2.HnswParams()
        params.m = self.m
        params.efConstruction = self.ef_construction
        params.ef = self.ef

        params.batchingParams.CopyFrom(self.batching_params._to_pb2())
        return params


class HnswSearchParams(object):
    def __init__(self, *, ef: Optional[int] = None) -> None:
        """
        Parameters for Hierarchical Navigable Small World (HNSW) search.

        HNSW is an algorithm used for approximate nearest neighbor search.

        Args:
            ef (Optional[int], optional): The parameter 'ef' controls the trade-off between search quality and search efficiency. It determines the size of the dynamic list of nearest neighbors (candidates) examined during the search phase. Larger values of 'ef' typically yield higher recall but slower search times. Defaults to None, meaning the algorithm uses a library-defined default value.

        Notes:
            - 'ef' stands for "exploration factor."
            - Setting 'ef' to a higher value increases the recall (i.e., the likelihood of finding the true nearest neighbors) at the cost of increased computational overhead during the search process.

        """
        self.ef = ef

    def _to_pb2(self):
        params = types_pb2.HnswSearchParams()
        params.ef = self.ef
        return params



import re
import threading
import warnings
import logging
from typing import Optional

import google.protobuf.empty_pb2
import grpc
import random

from . import types
from . import vector_db_pb2
from . import vector_db_pb2_grpc

empty = google.protobuf.empty_pb2.Empty()

logger = logging.getLogger(__name__)

class ChannelAndEndpoints(object):
    def __init__(
        self, channel: grpc.aio.Channel, endpoints: vector_db_pb2.ServerEndpointList
    ) -> None:
        self.channel = channel
        self.endpoints = endpoints


class VectorDbChannelProvider(object):
    """Proximus Channel Provider"""
    def __init__(
        self, seeds: tuple[types.HostPort, ...], listener_name: Optional[str] = None, is_loadbalancer: Optional[bool] = False
    ) -> None:
        if not seeds:
            raise Exception("at least one seed host needed")
        self._nodeChannels: dict[int, ChannelAndEndpoints] = {}
        self._seedChannels: dict[grpc.aio.Channel] = {}
        self._closed = False
        self._clusterId = 0
        self.seeds = seeds
        self.listener_name = listener_name
        self._is_loadbalancer = is_loadbalancer
        self._seedChannels = [
            self._create_channel_from_host_port(seed) for seed in self.seeds
        ]
        self._tend()

    async def close(self):
        self._closed = True
        for channel in self._seedChannels:
            await channel.close()

        for k, channelEndpoints in self._nodeChannels.items():
            if channelEndpoints.channel:
                await channelEndpoints.channel.close()

    def get_channel(self) -> grpc.Channel:
        if not self._is_loadbalancer:
            discovered_channels: list[ChannelAndEndpoints] = list(
                self._nodeChannels.values())
            if len(discovered_channels) <= 0:
                return self._seedChannels[0]


            # Return a random channel.
            channel = random.choice(discovered_channels).channel
            if channel:
                return channel

        return self._seedChannels[0]

    def _tend(self):
        if self._is_loadbalancer:
            # Skip tend if we are behind a load-balancer
            return

        # TODO: Worry about thread safety
        temp_endpoints: dict[int, vector_db_pb2.ServerEndpointList] = {}

        if self._closed:
            return

        try:
            update_endpoints = False
            channels = self._seedChannels + [
                x.channel for x in self._nodeChannels.values()
            ]
            for seedChannel in channels:
                try:
                    stub = vector_db_pb2_grpc.ClusterInfoStub(seedChannel)
                    newClusterId = stub.GetClusterId(empty).id

                    if newClusterId == self._clusterId:
                        continue

                    update_endpoints = True
                    self._clusterId = newClusterId
                    endpoints = stub.GetClusterEndpoints(
                        vector_db_pb2.ClusterNodeEndpointsRequest(
                            listenerName=self.listener_name
                        )
                    ).endpoints

                    if len(endpoints) > len(temp_endpoints):
                        temp_endpoints = endpoints

                except Exception as e:
                    logger.debug("failure tending cluster endpoints: " + str(e))

            if update_endpoints:
                for node, newEndpoints in temp_endpoints.items():
                    channel_endpoints = self._nodeChannels.get(node)
                    add_new_channel = True
                    if channel_endpoints:
                        # We have this node. Check if the endpoints changed.
                        if channel_endpoints.endpoints == newEndpoints:
                            # Nothing to be done for this node
                            add_new_channel = False
                        else:
                            # TODO: Wait for all calls to drain
                            channel_endpoints.channel.close()
                            add_new_channel = True

                    if add_new_channel:
                        # We have discovered a new node
                        new_channel = self._create_channel_from_server_endpoint_list(
                            newEndpoints
                        )
                        self._nodeChannels[node] = ChannelAndEndpoints(
                            new_channel, newEndpoints
                        )

                for node, channel_endpoints in self._nodeChannels.items():
                    if not temp_endpoints.get(node):
                        # TODO: Wait for all calls to drain
                        channel_endpoints.channel.close()
                        del self._nodeChannels[node]

        except Exception as e:
            logger.debug("failure tending: " + str(e))

        if not self._closed:
            # TODO: check tend interval.
            threading.Timer(1, self._tend).start()

    def _create_channel_from_host_port(self, host: types.HostPort) -> grpc.aio.Channel:
        return self._create_channel(host.host, host.port, host.isTls)

    def _create_channel_from_server_endpoint_list(
        self, endpoints: vector_db_pb2.ServerEndpointList
    ) -> grpc.aio.Channel:
        # TODO: Create channel with all endpoints
        for endpoint in endpoints.endpoints:
            if ":" in endpoint.address:
                # TODO: Ignoring IPv6 for now. Needs fix
                continue
            try:
                return self._create_channel(
                    endpoint.address, endpoint.port, endpoint.isTls
                )
            except Exception as e:
                logger.debug("failure creating channel: " + str(e))

    def _create_channel(self, host: str, port: int, isTls: bool) -> grpc.aio.Channel:
        # TODO: Take care of TLS
        host = re.sub(r"%.*", "", host)
        return grpc.aio.insecure_channel(f"{host}:{port}")


from typing import Any

from . import types
from . import types_pb2


def toVectorDbValue(value: Any) -> types_pb2.Value:
    if isinstance(value, str):
        return types_pb2.Value(stringValue=value)
    elif isinstance(value, int):
        return types_pb2.Value(longValue=value)
    elif isinstance(value, float):
        return types_pb2.Value(doubleValue=value)
    elif isinstance(value, (bytes, bytearray)):
        return types_pb2.Value(bytesValue=value)
    elif isinstance(value, list) and value:
        # TODO: Convert every element correctly to destination type.
        if isinstance(value[0], float):
            return types_pb2.Value(
                vectorValue=types_pb2.Vector(
                    floatData={"value": [float(x) for x in value]}
                )
            )
        elif isinstance(value[0], bool):
            return types_pb2.Value(
                vectorValue=types_pb2.Vector(
                    boolData={"value": [True if x else False for x in value]}
                )
            )
        else:
            return types_pb2.Value(
                listValue=types_pb2.List(entries=[toVectorDbValue(x) for x in value])
            )
    elif isinstance(value, dict):
        d = types_pb2.Value(
            mapValue=types_pb2.Map(
                entries=[
                    types_pb2.MapEntry(key=toMapKey(k), value=toVectorDbValue(v))
                    for k, v in value.items()
                ]
            )
        )
        return d
    else:
        raise Exception("Invalid type " + str(type(value)))


def toMapKey(value):
    if isinstance(value, str):
        return types_pb2.MapKey(stringValue=value)
    elif isinstance(value, int):
        return types_pb2.MapKey(longValue=value)
    elif isinstance(value, (bytes, bytearray)):
        return types_pb2.MapKey(bytesValue=value)
    elif isinstance(value, float):
        return types_pb2.MapKey(doubleValue=value)
    else:
        raise Exception("Invalid map key type " + str(type(value)))


def fromVectorDbKey(key: types_pb2.Key) -> types.Key:
    keyValue = None
    if key.HasField("stringValue"):
        keyValue = key.stringValue
    elif key.HasField("intValue"):
        keyValue = key.intValue
    elif key.HasField("longValue"):
        keyValue = key.longValue
    elif key.HasField("bytesValue"):
        keyValue = key.bytesValue

    return types.Key(
        namespace=key.namespace, set=key.set, digest=key.digest, key=keyValue
    )


def fromVectorDbRecord(record: types_pb2.Record) -> dict[str, Any]:
    bins = {}
    for bin in record.bins:
        bins[bin.name] = fromVectorDbValue(bin.value)

    return bins


def fromVectorDbNeighbor(input: types_pb2.Neighbor) -> types.Neighbor:
    return types.Neighbor(
        key=fromVectorDbKey(input.key),
        bins=fromVectorDbRecord(input.record),
        distance=input.distance,
    )


def fromVectorDbValue(input: types_pb2.Value) -> Any:
    if input.HasField("stringValue"):
        return input.stringValue
    elif input.HasField("intValue"):
        return input.intValue
    elif input.HasField("longValue"):
        return input.longValue
    elif input.HasField("bytesValue"):
        return input.bytesValue
    elif input.HasField("mapValue"):
        dict = {}
        for entry in input.mapValue.entries:
            k = fromVectorDbValue(entry.key)
            v = fromVectorDbValue(entry.value)
            dict[k] = v
        return dict
    elif input.HasField("listValue"):
        return [fromVectorDbValue(v) for v in input.listValue.entries]
    elif input.HasField("vectorValue"):
        vector = input.vectorValue
        if vector.HasField("floatData"):
            return [v for v in vector.floatData.value]
        if vector.HasField("boolData"):
            return [v for v in vector.boolData.value]

    return None



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: auth.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\nauth.proto\x12\x10\x61\x65rospike.vector\":\n\x14\x41\x65rospikeAuthRequest\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x10\n\x08password\x18\x02 \x01(\t\"&\n\x15\x41\x65rospikeAuthResponse\x12\r\n\x05token\x18\x01 \x01(\t2g\n\x0b\x41uthService\x12X\n\x03Get\x12&.aerospike.vector.AerospikeAuthRequest\x1a\'.aerospike.vector.AerospikeAuthResponse\"\x00\x42=\n\x1b\x63om.aerospike.vector.clientP\x01Z\x1c\x61\x65rospike.com/vector/protos/b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'auth_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\033com.aerospike.vector.clientP\001Z\034aerospike.com/vector/protos/'
  _globals['_AEROSPIKEAUTHREQUEST']._serialized_start=32
  _globals['_AEROSPIKEAUTHREQUEST']._serialized_end=90
  _globals['_AEROSPIKEAUTHRESPONSE']._serialized_start=92
  _globals['_AEROSPIKEAUTHRESPONSE']._serialized_end=130
  _globals['_AUTHSERVICE']._serialized_start=132
  _globals['_AUTHSERVICE']._serialized_end=235
# @@protoc_insertion_point(module_scope)



import asyncio
import logging
import sys
import time
from typing import Any, Optional, Union

import google.protobuf.empty_pb2
from google.protobuf.json_format import MessageToDict
import grpc

from . import index_pb2
from . import index_pb2_grpc
from . import types
from . import types_pb2
from . import vectordb_channel_provider

empty = google.protobuf.empty_pb2.Empty()
logger = logging.getLogger(__name__)


class VectorDbAdminClient(object):
    """
    Aerospike Vector Admin Client

    This client is designed to conduct Aerospike Vector administrative operation such as creating indexes, querying index information, and dropping indexes.
    """

    def __init__(
        self,
        *,
        seeds: Union[types.HostPort, tuple[types.HostPort, ...]],
        listener_name: Optional[str] = None,
        is_loadbalancer: Optional[bool] = False,
    ) -> None:

        if not seeds:
            raise Exception("at least one seed host needed")

        if isinstance(seeds, types.HostPort):
            seeds = (seeds,)

        self._channelProvider = vectordb_channel_provider.VectorDbChannelProvider(
            seeds, listener_name, is_loadbalancer
        )
        """
        Initialize the Aerospike Vector Admin Client.

        Args:
            seeds (Union[types.HostPort, tuple[types.HostPort, ...]]):
                Used to create appropriate gRPC channels for interacting with Aerospike Vector.
            listener_name (Optional[str], optional):
                Advertised listener for the client. Defaults to None.
            is_loadbalancer (bool, optional):
                If true, the first seed address will be treated as a load balancer node.

        Raises:
            Exception: Raised when no seed host is provided.
        """

    async def index_create(
        self,
        *,
        namespace: str,
        name: str,
        vector_field: str,
        dimensions: int,
        vector_distance_metric: Optional[types.VectorDistanceMetric] = (
            types.VectorDistanceMetric.SQUARED_EUCLIDEAN
        ),
        sets: Optional[str] = None,
        index_params: Optional[types.HnswParams] = None,
        index_meta_data: Optional[dict[str, str]] = None,
    ) -> None:
        """
        Create an index.

        Args:
            namespace (str): The namespace for the index.
            name (str): The name of the index.
            vector_field (str): The name of the field containing vector data.
            dimensions (int): The number of dimensions in the vector data.
            vector_distance_metric (Optional[types.VectorDistanceMetric], optional):
                The distance metric used to compare when performing a vector search.
                Defaults to VectorDistanceMetric.SQUARED_EUCLIDEAN.
            sets (Optional[str], optional): The set used for the index. Defaults to None.
            index_params (Optional[types.HnswParams], optional):
                Parameters used for tuning vector search. Defaults to None. If index_params is None, then the default values specified for
                types.HnswParams will be used.
            index_meta_data (Optional[dict[str, str]], optional): Meta data associated with the index. Defaults to None.

        Raises:
            grpc.RpcError: Raised if an error occurs during the RPC communication with the server while attempting to create the index.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.

        Note:
            This method creates an index with the specified parameters and waits for the index creation to complete.
            It waits for up to 100,000 seconds for the index creation to complete.
        """
        index_stub = index_pb2_grpc.IndexServiceStub(
            self._channelProvider.get_channel()
        )
        if sets and not sets.strip():
            sets = None

        logger.debug(
            "Creating index: namespace=%s, name=%s, vector_field=%s, dimensions=%d, vector_distance_metric=%s, "
            "sets=%s, index_params=%s, index_meta_data=%s",
            namespace,
            name,
            vector_field,
            dimensions,
            vector_distance_metric,
            sets,
            index_params,
            index_meta_data,
        )
        if index_params != None:
            index_params = index_params._to_pb2()
        try:
            await index_stub.Create(
                types_pb2.IndexDefinition(
                    id=types_pb2.IndexId(namespace=namespace, name=name),
                    vectorDistanceMetric=vector_distance_metric.value,
                    setFilter=sets,
                    hnswParams=index_params,
                    bin=vector_field,
                    dimensions=dimensions,
                    labels=index_meta_data,
                )
            )
        except grpc.RpcError as e:
            logger.error("Failed with error: %s", e)
            raise e
        try:
            await self._wait_for_index_creation(
                namespace=namespace, name=name, timeout=100_000
            )
        except grpc.RpcError as e:
            logger.error("Failed waiting for creation with error: %s", e)
            raise e

    async def index_drop(self, *, namespace: str, name: str) -> None:
        index_stub = index_pb2_grpc.IndexServiceStub(
            self._channelProvider.get_channel()
        )
        """
        Drop an index.

        Args:
            namespace (str): The namespace of the index.
            name (str): The name of the index.

        Raises:
            grpc.RpcError: Raised if an error occurs during the RPC communication with the server while attempting to create the index.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.

        Note:
            This method drops an index with the specified parameters and waits for the index deletion to complete.
            It waits for up to 100,000 seconds for the index deletion to complete.
        """
        logger.debug("Dropping index: namespace=%s, name=%s", namespace, name)
        try:
            await index_stub.Drop(types_pb2.IndexId(namespace=namespace, name=name))
        except grpc.RpcError as e:
            logger.error("Failed with error: %s", e)
            raise e
        try:
            await self._wait_for_index_deletion(
                namespace=namespace, name=name, timeout=100_000
            )
        except grpc.RpcError as e:
            logger.error("Failed waiting for deletion with error: %s", e)
            raise e

    async def index_list(self) -> list[dict]:
        """
        List all indices.

        Returns:
            list[dict]: A list of indices.

        Raises:
            grpc.RpcError: Raised if an error occurs during the RPC communication with the server while attempting to create the index.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.
        """

        index_stub = index_pb2_grpc.IndexServiceStub(
            self._channelProvider.get_channel()
        )

        logger.debug("Getting index list")
        try:
            response = await index_stub.List(empty)
        except grpc.RpcError as e:
            logger.error("Failed with error: %s", e)
            raise e
        response_list = []
        for index in response.indices:
            response_dict = MessageToDict(index)

            # Modifying dict to adhere to PEP-8 naming
            hnsw_params_dict = response_dict.pop("hnswParams", None)

            hnsw_params_dict["ef_construction"] = hnsw_params_dict.pop(
                "efConstruction", None
            )

            batching_params_dict = hnsw_params_dict.pop("batchingParams", None)
            batching_params_dict["max_records"] = batching_params_dict.pop(
                "maxRecords", None
            )
            hnsw_params_dict["batching_params"] = batching_params_dict

            response_dict["hnsw_params"] = hnsw_params_dict
            response_list.append(response_dict)
        return response_list

    async def index_get(
        self, *, namespace: str, name: str
    ) -> dict[str, Union[int, str]]:
        """
        Retrieve the information related with an index.

        Args:
            namespace (str): The namespace of the index.
            name (str): The name of the index.

        Returns:
            dict[str, Union[int, str]: Information about an index.

        Raises:
            grpc.RpcError: Raised if an error occurs during the RPC communication with the server while attempting to create the index.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.

        """

        index_stub = index_pb2_grpc.IndexServiceStub(
            self._channelProvider.get_channel()
        )

        logger.debug(
            "Getting index information: namespace=%s, name=%s", namespace, name
        )
        try:
            response = await index_stub.Get(
                types_pb2.IndexId(namespace=namespace, name=name)
            )
        except grpc.RpcError as e:
            logger.error("Failed with error: %s", e)
            raise e

        response_dict = MessageToDict(response)

        # Modifying dict to adhere to PEP-8 naming
        hnsw_params_dict = response_dict.pop("hnswParams", None)

        hnsw_params_dict["ef_construction"] = hnsw_params_dict.pop(
            "efConstruction", None
        )

        batching_params_dict = hnsw_params_dict.pop("batchingParams", None)
        batching_params_dict["max_records"] = batching_params_dict.pop(
            "maxRecords", None
        )
        hnsw_params_dict["batching_params"] = batching_params_dict

        response_dict["hnsw_params"] = hnsw_params_dict
        return response_dict

    async def index_get_status(self, *, namespace: str, name: str) -> int:
        """
        Retrieve the number of records queued to be merged into an index.

        Args:
            namespace (str): The namespace of the index.
            name (str): The name of the index.

        Returns:
            int: Records queued to be merged into an index.

        Raises:
            grpc.RpcError: Raised if an error occurs during the RPC communication with the server while attempting to create the index.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.

        Note:
            This method retrieves the status of the specified index. If index_get_status is called the vector client puts some records into Aerospike Vector,
            the records may not immediately begin to merge into the index. To wait for all records to be merged into an index, use vector_client.wait_for_index_completion.

            Warning: This API is subject to change.
        """
        index_stub = index_pb2_grpc.IndexServiceStub(
            self._channelProvider.get_channel()
        )
        logger.debug("Getting index status: namespace=%s, name=%s", namespace, name)
        try:
            response = await index_stub.GetStatus(
                types_pb2.IndexId(namespace=namespace, name=name)
            )
        except grpc.RpcError as e:
            logger.error("Failed with error: %s", e)
            raise e

        return response.unmergedRecordCount

    async def _wait_for_index_creation(
        self, *, namespace: str, name: str, timeout: int = sys.maxsize
    ) -> None:
        """
        Wait for the index to be created.
        """

        # Wait interval between polling
        index_stub = index_pb2_grpc.IndexServiceStub(
            self._channelProvider.get_channel()
        )
        wait_interval = 0.100

        start_time = time.monotonic()
        while True:
            if start_time + timeout < time.monotonic():
                raise "timed-out waiting for index creation"
            try:
                await index_stub.GetStatus(
                    types_pb2.IndexId(namespace=namespace, name=name)
                )
                logger.debug("Index created succesfully")
                # Index has been created
                return
            except grpc.RpcError as e:
                if e.code() in (grpc.StatusCode.UNAVAILABLE, grpc.StatusCode.NOT_FOUND):

                    # Wait for some more time.
                    await asyncio.sleep(wait_interval)
                else:
                    logger.error("Failed with error: %s", e)
                    raise e

    async def _wait_for_index_deletion(
        self, *, namespace: str, name: str, timeout: int = sys.maxsize
    ) -> None:
        """
        Wait for the index to be deleted.
        """

        # Wait interval between polling
        index_stub = index_pb2_grpc.IndexServiceStub(
            self._channelProvider.get_channel()
        )
        wait_interval = 0.100

        start_time = time.monotonic()
        while True:
            if start_time + timeout < time.monotonic():
                raise "timed-out waiting for index creation"
            try:
                await index_stub.GetStatus(
                    types_pb2.IndexId(namespace=namespace, name=name)
                )
                # Wait for some more time.
                await asyncio.sleep(wait_interval)
            except grpc.RpcError as e:
                if e.code() in (grpc.StatusCode.UNAVAILABLE, grpc.StatusCode.NOT_FOUND):
                    logger.debug("Index deleted succesfully")
                    # Index has been created
                    return
                else:
                    raise e

    async def close(self):
        """
        Close the Aerospike Vector Admin Client.

        This method closes gRPC channels connected to Aerospike Vector.

        Note:
            This method should be called when the VectorDbAdminClient is no longer needed to release resources.
        """
        await self._channelProvider.close()

    async def __aenter__(self):
        """
        Enter an asynchronous context manager for the admin client.

        Returns:
            VectorDbAdminlient: Aerospike Vector Admin Client instance.
        """
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Exit an asynchronous context manager for the admin client.
        """
        await self.close()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: types.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0btypes.proto\x12\x10\x61\x65rospike.vector\"\xa1\x01\n\x03Key\x12\x11\n\tnamespace\x18\x01 \x01(\t\x12\x10\n\x03set\x18\x02 \x01(\tH\x01\x88\x01\x01\x12\x0e\n\x06\x64igest\x18\x03 \x01(\x0c\x12\x15\n\x0bstringValue\x18\x04 \x01(\tH\x00\x12\x14\n\nbytesValue\x18\x05 \x01(\x0cH\x00\x12\x12\n\x08intValue\x18\x06 \x01(\x05H\x00\x12\x13\n\tlongValue\x18\x07 \x01(\x03H\x00\x42\x07\n\x05valueB\x06\n\x04_set\"\x19\n\x08\x42oolData\x12\r\n\x05value\x18\x01 \x03(\x08\"\x1a\n\tFloatData\x12\r\n\x05value\x18\x01 \x03(\x02\"\x94\x01\n\x06MapKey\x12\x15\n\x0bstringValue\x18\x01 \x01(\tH\x00\x12\x14\n\nbytesValue\x18\x02 \x01(\x0cH\x00\x12\x12\n\x08intValue\x18\x03 \x01(\x05H\x00\x12\x13\n\tlongValue\x18\x04 \x01(\x03H\x00\x12\x14\n\nfloatValue\x18\x05 \x01(\x02H\x00\x12\x15\n\x0b\x64oubleValue\x18\x06 \x01(\x01H\x00\x42\x07\n\x05value\"Y\n\x08MapEntry\x12%\n\x03key\x18\x01 \x01(\x0b\x32\x18.aerospike.vector.MapKey\x12&\n\x05value\x18\x02 \x01(\x0b\x32\x17.aerospike.vector.Value\"2\n\x03Map\x12+\n\x07\x65ntries\x18\x01 \x03(\x0b\x32\x1a.aerospike.vector.MapEntry\"0\n\x04List\x12(\n\x07\x65ntries\x18\x01 \x03(\x0b\x32\x17.aerospike.vector.Value\"r\n\x06Vector\x12.\n\x08\x62oolData\x18\x01 \x01(\x0b\x32\x1a.aerospike.vector.BoolDataH\x00\x12\x30\n\tfloatData\x18\x02 \x01(\x0b\x32\x1b.aerospike.vector.FloatDataH\x00\x42\x06\n\x04\x64\x61ta\"\x9c\x02\n\x05Value\x12\x15\n\x0bstringValue\x18\x01 \x01(\tH\x00\x12\x14\n\nbytesValue\x18\x02 \x01(\x0cH\x00\x12\x12\n\x08intValue\x18\x03 \x01(\x05H\x00\x12\x13\n\tlongValue\x18\x04 \x01(\x03H\x00\x12\x14\n\nfloatValue\x18\x05 \x01(\x02H\x00\x12\x15\n\x0b\x64oubleValue\x18\x06 \x01(\x01H\x00\x12)\n\x08mapValue\x18\x07 \x01(\x0b\x32\x15.aerospike.vector.MapH\x00\x12+\n\tlistValue\x18\x08 \x01(\x0b\x32\x16.aerospike.vector.ListH\x00\x12/\n\x0bvectorValue\x18\t \x01(\x0b\x32\x18.aerospike.vector.VectorH\x00\x42\x07\n\x05value\";\n\x03\x42in\x12\x0c\n\x04name\x18\x01 \x01(\t\x12&\n\x05value\x18\x02 \x01(\x0b\x32\x17.aerospike.vector.Value\"U\n\x06Record\x12\x12\n\ngeneration\x18\x01 \x01(\r\x12\x12\n\nexpiration\x18\x02 \x01(\r\x12#\n\x04\x62ins\x18\x03 \x03(\x0b\x32\x15.aerospike.vector.Bin\"z\n\x08Neighbor\x12\"\n\x03key\x18\x01 \x01(\x0b\x32\x15.aerospike.vector.Key\x12-\n\x06record\x18\x02 \x01(\x0b\x32\x18.aerospike.vector.RecordH\x00\x88\x01\x01\x12\x10\n\x08\x64istance\x18\x03 \x01(\x02\x42\t\n\x07_record\"*\n\x07IndexId\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tnamespace\x18\x02 \x01(\t\"\xa8\x01\n\nHnswParams\x12\x0e\n\x01m\x18\x01 \x01(\rH\x00\x88\x01\x01\x12\x1b\n\x0e\x65\x66\x43onstruction\x18\x02 \x01(\rH\x01\x88\x01\x01\x12\x0f\n\x02\x65\x66\x18\x03 \x01(\rH\x02\x88\x01\x01\x12<\n\x0e\x62\x61tchingParams\x18\x04 \x01(\x0b\x32$.aerospike.vector.HnswBatchingParamsB\x04\n\x02_mB\x11\n\x0f_efConstructionB\x05\n\x03_ef\"*\n\x10HnswSearchParams\x12\x0f\n\x02\x65\x66\x18\x01 \x01(\rH\x00\x88\x01\x01\x42\x05\n\x03_ef\"\x84\x01\n\x12HnswBatchingParams\x12\x17\n\nmaxRecords\x18\x01 \x01(\rH\x00\x88\x01\x01\x12\x15\n\x08interval\x18\x02 \x01(\rH\x01\x88\x01\x01\x12\x15\n\x08\x64isabled\x18\x03 \x01(\x08H\x02\x88\x01\x01\x42\r\n\x0b_maxRecordsB\x0b\n\t_intervalB\x0b\n\t_disabled\"W\n\x15\x41\x65rospikeIndexStorage\x12\x16\n\tnamespace\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x10\n\x03set\x18\x02 \x01(\tH\x01\x88\x01\x01\x42\x0c\n\n_namespaceB\x06\n\x04_set\"\xec\x03\n\x0fIndexDefinition\x12%\n\x02id\x18\x01 \x01(\x0b\x32\x19.aerospike.vector.IndexId\x12)\n\x04type\x18\x02 \x01(\x0e\x32\x1b.aerospike.vector.IndexType\x12\x12\n\ndimensions\x18\x03 \x01(\r\x12\x44\n\x14vectorDistanceMetric\x18\x04 \x01(\x0e\x32&.aerospike.vector.VectorDistanceMetric\x12\x0b\n\x03\x62in\x18\x05 \x01(\t\x12\x16\n\tsetFilter\x18\x06 \x01(\tH\x02\x88\x01\x01\x12\x32\n\nhnswParams\x18\x07 \x01(\x0b\x32\x1c.aerospike.vector.HnswParamsH\x00\x12\x43\n\x10\x61\x65rospikeStorage\x18\x08 \x01(\x0b\x32\'.aerospike.vector.AerospikeIndexStorageH\x01\x12=\n\x06labels\x18\t \x03(\x0b\x32-.aerospike.vector.IndexDefinition.LabelsEntry\x1a-\n\x0bLabelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x08\n\x06paramsB\t\n\x07storageB\x0c\n\n_setFilter\"I\n\x13IndexDefinitionList\x12\x32\n\x07indices\x18\x01 \x03(\x0b\x32!.aerospike.vector.IndexDefinition\"\x18\n\x07\x42oolean\x12\r\n\x05value\x18\x01 \x01(\x08*f\n\x14VectorDistanceMetric\x12\x15\n\x11SQUARED_EUCLIDEAN\x10\x00\x12\n\n\x06\x43OSINE\x10\x01\x12\x0f\n\x0b\x44OT_PRODUCT\x10\x02\x12\r\n\tMANHATTAN\x10\x03\x12\x0b\n\x07HAMMING\x10\x04*\x15\n\tIndexType\x12\x08\n\x04HNSW\x10\x00\x42=\n\x1b\x63om.aerospike.vector.clientP\x01Z\x1c\x61\x65rospike.com/vector/protos/b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'types_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\033com.aerospike.vector.clientP\001Z\034aerospike.com/vector/protos/'
  _globals['_INDEXDEFINITION_LABELSENTRY']._options = None
  _globals['_INDEXDEFINITION_LABELSENTRY']._serialized_options = b'8\001'
  _globals['_VECTORDISTANCEMETRIC']._serialized_start=2350
  _globals['_VECTORDISTANCEMETRIC']._serialized_end=2452
  _globals['_INDEXTYPE']._serialized_start=2454
  _globals['_INDEXTYPE']._serialized_end=2475
  _globals['_KEY']._serialized_start=34
  _globals['_KEY']._serialized_end=195
  _globals['_BOOLDATA']._serialized_start=197
  _globals['_BOOLDATA']._serialized_end=222
  _globals['_FLOATDATA']._serialized_start=224
  _globals['_FLOATDATA']._serialized_end=250
  _globals['_MAPKEY']._serialized_start=253
  _globals['_MAPKEY']._serialized_end=401
  _globals['_MAPENTRY']._serialized_start=403
  _globals['_MAPENTRY']._serialized_end=492
  _globals['_MAP']._serialized_start=494
  _globals['_MAP']._serialized_end=544
  _globals['_LIST']._serialized_start=546
  _globals['_LIST']._serialized_end=594
  _globals['_VECTOR']._serialized_start=596
  _globals['_VECTOR']._serialized_end=710
  _globals['_VALUE']._serialized_start=713
  _globals['_VALUE']._serialized_end=997
  _globals['_BIN']._serialized_start=999
  _globals['_BIN']._serialized_end=1058
  _globals['_RECORD']._serialized_start=1060
  _globals['_RECORD']._serialized_end=1145
  _globals['_NEIGHBOR']._serialized_start=1147
  _globals['_NEIGHBOR']._serialized_end=1269
  _globals['_INDEXID']._serialized_start=1271
  _globals['_INDEXID']._serialized_end=1313
  _globals['_HNSWPARAMS']._serialized_start=1316
  _globals['_HNSWPARAMS']._serialized_end=1484
  _globals['_HNSWSEARCHPARAMS']._serialized_start=1486
  _globals['_HNSWSEARCHPARAMS']._serialized_end=1528
  _globals['_HNSWBATCHINGPARAMS']._serialized_start=1531
  _globals['_HNSWBATCHINGPARAMS']._serialized_end=1663
  _globals['_AEROSPIKEINDEXSTORAGE']._serialized_start=1665
  _globals['_AEROSPIKEINDEXSTORAGE']._serialized_end=1752
  _globals['_INDEXDEFINITION']._serialized_start=1755
  _globals['_INDEXDEFINITION']._serialized_end=2247
  _globals['_INDEXDEFINITION_LABELSENTRY']._serialized_start=2167
  _globals['_INDEXDEFINITION_LABELSENTRY']._serialized_end=2212
  _globals['_INDEXDEFINITIONLIST']._serialized_start=2249
  _globals['_INDEXDEFINITIONLIST']._serialized_end=2322
  _globals['_BOOLEAN']._serialized_start=2324
  _globals['_BOOLEAN']._serialized_end=2348
# @@protoc_insertion_point(module_scope)
